#!/usr/bin/env python
# -*- coding: utf-8 -*-

from uscpi.client import ClientBase


class Instrument:
    """Instrument representation.

    Implements common commands associated with IEEE-488.2.
    """

    def __init__(self, client: ClientBase) -> None:
        self.client = client

    @property
    def client(self) -> ClientBase:
        return self._client

    @client.setter
    def client(self, value: ClientBase) -> None:
        if isinstance(value, ClientBase):
            self._client = value
            return
        raise TypeError("Expected a ClientBase object")

    def cls(self) -> None:
        """Clear Status Command.

        Clears the event registers in all register groups.
        Also clears the error queue.
        """

        return self.client.write(b"*CLS\n")

    def ese(self, value: int | None = None) -> bytes | None:
        """Event Status Enable Command and Query.

        Enables bits in the enable register for the Standard
        Event Register group. The selected bits are then
        reported to bit 5 of the Status Byte Register. An
        enable register defines which bits in the event
        register will be reported to the Status Byte
        register group. You can write to or read from an
        enable register.
        """

        if isinstance(value, int):
            if value not in range(256):
                raise ValueError(value)
            return self.client.write(f"*ESE {value}\n".encode())
        return self.client.write_readline(b"*ESE?\n")

    def esr(self) -> bytes:
        """Standard Event Status Register Query.

        Queries the event register for the Standard Event
        Register group.

        An event register is a read-only register that
        latches events from the condition register. While an
        event bit is set, subsequent events corresponding to
        that bit are ignored.
        """

        return self.client.write_readline(b"*ESR?\n")

    def idn(self) -> bytes:
        """Identification Query.

        Returns the instrument’s identification string.
        """

        return self.client.write_readline(b"*IDN?\n")

    def opc(self, complete: bool = False) -> bytes | None:
        """Operation Complete Command and Query.

        Sets "Operation Complete" (bit 0) in the Standard
        Event register at the completion of the current
        operation.
        """

        if isinstance(complete, bool) and complete:
            return self.client.write(b"*OPC\n")
        return self.client.write_readline(b"*OPC?\n")

    def rst(self) -> None:
        """Reset Command.

        Resets instrument to factory default state,
        independent of MMEMory:STATe:RECall:AUTO setting.
        This is similar to SYSTem:PRESet. The difference is
        that *RST resets the instrument for SCPI operation,
        and SYSTem:PRESet resets the instrument for
        front-panel operation. As a result, *RST turns the
        histogram and statistics off, and SYSTem:PRESet
        turns them on (CALC:TRAN:HIST:STAT ON).
        """

        return self.client.write(b"*RST\n")

    def sre(self, value: int | None = None) -> bytes | None:
        """Service Request Enable Command and Query.

        Service Request Enable. Enables bits in the enable
        register for the Status Byte Register group. An
        enable register defines which bits in the event
        register will be reported to the Status Byte
        register group. You can write to or read from an
        enable register.

        If a value is not specified, returns the content
        of the Service Request Enable register.
        """

        if isinstance(value, int):
            if value not in range(256):
                raise ValueError(value)
            return self.client.write(f"*SRE {value}\n".encode())
        return self.client.write_readline(b"*SRE?\n")

    def stb(self) -> bytes:
        """Read Status Byte Query.

        Queries the condition register for the Status Byte
        Register group and returns a decimal value equal to
        the binary-weighted sum of all bits set in the
        register.

        A condition register continuously monitors the state
        of the instrument. Condition register bits are
        updated in real time; they are neither latched nor
        buffered.
        """

        return self.client.write_readline(b"*STB?\n")

    def trg(self) -> None:
        """Trigger Command.

        Triggers the instrument if TRIGger:SOURce BUS is
        selected.
        """

        return self.client.write("*TRG\n")

    def tst(self) -> bytes:
        """Self-Test Query.

        Performs a basic self-test of the instrument and
        returns a pass/fail indication. The TEST:ALL?
        self-test is more comprehensive than the *TST?
        self-test.

        Returns an integer value in the range of -32767
        and 32767. A response of 0 indicates that the
        self-test completed without errors detected.
        """

        return self.client.write_readline(b"*TST?\n")

    def wai(self) -> None:
        """Wait-to-Continue Command.

        Configures the instrument's output buffer to wait
        for all pending operations to complete before
        executing any additional commands over the
        interface.
        """

        return self.client.write(b"*WAI\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import IsolatedAsyncioTestCase
from unittest import main
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import patch

from uscpi.instrument import Instrument


@patch("uscpi.client.ClientBase", autospec=True)
class TestInstrument(IsolatedAsyncioTestCase):
    async def test_cls(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.cls()
        mock_client_base.write.assert_awaited()

    async def test_ese_write_readline(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.ese(), b"test")

    async def test_ese_write(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.ese(128)
        mock_client_base.write.assert_awaited()

    async def test_esr(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.esr(), b"test")

    async def test_idn(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.idn(), b"test")

    async def test_opc_write_readline(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.opc(), b"test")

    async def test_opc_write(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.opc(complete=True)
        mock_client_base.write.assert_awaited()

    async def test_rst(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.rst()
        mock_client_base.write.assert_awaited()

    async def test_sre_write_readline(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.sre(), b"test")

    async def test_sre_write(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.sre(value=128)
        mock_client_base.write.assert_awaited()

    async def test_stb(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.stb(), b"test")

    async def test_trg(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.trg()
        mock_client_base.write.assert_awaited()

    async def test_tst(self, mock_client_base):
        mock_client_base.write_readline = AsyncMock()
        mock_client_base.write_readline.return_value = b"test"
        instrument = Instrument(mock_client_base)
        self.assertEqual(await instrument.tst(), b"test")

    async def test_wai(self, mock_client_base):
        mock_client_base.write = AsyncMock()
        instrument = Instrument(mock_client_base)
        await instrument.wai()
        mock_client_base.write.assert_awaited()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC
from abc import abstractmethod
from asyncio import get_running_loop
from asyncio import Lock
from asyncio import StreamReader
from asyncio import StreamWriter
from asyncio import wait_for
from functools import wraps
from typing import Callable

from uscpi.protocol import CallbackProtocol


async def open_connection(
    host: str,
    port: int,
    limit: int | float,
    connection_made_cb: Callable,
    connection_lost_cb: Callable,
    data_received_cb: Callable,
    eof_received_cb: Callable,
    **kwargs,
) -> tuple:
    """Open connection.

    Identical to the CPython helper method implementation
    but with optional callback arguments.
    """

    loop = get_running_loop()
    reader = StreamReader(limit=limit, loop=loop)
    protocol = CallbackProtocol(reader, loop=loop)
    protocol.connection_made_cb = connection_made_cb
    protocol.connection_lost_cb = connection_lost_cb
    protocol.data_received_cb = data_received_cb
    protocol.eof_received_cb = eof_received_cb
    transport, _ = await loop.create_connection(lambda: protocol, host, port, **kwargs)
    writer = StreamWriter(transport, protocol, reader, loop)
    return reader, writer


class ClientBase(ABC):
    """
    Client base representation.
    """

    @abstractmethod
    async def close(self, *args, **kwargs) -> None:
        ...

    @abstractmethod
    async def open(self, *args, **kwargs) -> None:
        ...


class TCP(ClientBase):
    """
    TCP client representation.
    """

    lock: Lock | None = None
    reader: StreamReader | None = None
    writer: StreamWriter | None = None

    def __init__(
        self,
        host: str,
        port: int,
        timeout: int | float = 2**16,
        connection_made_cb: Callable = None,
        connection_lost_cb: Callable = None,
        data_received_cb: Callable = None,
        eof_received_cb: Callable = None,
    ) -> None:
        self.host = host
        self.port = port
        self.timeout = timeout
        self.connection_made_cb = connection_made_cb
        self.connection_lost_cb = connection_lost_cb
        self.data_received_cb = data_received_cb
        self.eof_received_cb = eof_received_cb

    @staticmethod
    def connection(func):
        """
        Attempt to establish/re-establish connection.
        """

        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            if not isinstance(self.lock, Lock):
                self.lock = Lock()
            async with self.lock:
                if not self.is_connected():
                    await self.open()
            coroutine = func(self, *args, **kwargs)
            return await coroutine

        return wrapper

    def is_eof(self) -> bool:
        """
        Check if EOF stream reader flag reached.
        """

        reader_exists = isinstance(self.reader, StreamReader)
        return reader_exists and self.reader.at_eof()

    def is_connected(self) -> bool:
        """
        Check if stream reader exists.
        """

        reader_exists = isinstance(self.reader, StreamReader)
        return reader_exists and not self.is_eof()

    async def close(self, *args, **kwargs) -> None:
        """
        Close active reader and writer streams.
        """

        if isinstance(self.writer, StreamWriter):
            self.writer.close()
            await self.writer.wait_closed()
        self.reader = None
        self.writer = None

    async def open(self, *args, **kwargs) -> None:
        """
        Open new reader and writer streams.
        """

        if self.is_connected():
            raise Exception("Already connected.")
        await self.close()
        coroutine = open_connection(
            self.host,
            self.port,
            self.timeout,
            self.connection_made_cb,
            self.connection_lost_cb,
            self.data_received_cb,
            self.eof_received_cb,
        )
        if isinstance(self.timeout, (int, float)):
            coroutine = wait_for(coroutine, self.timeout)
        self.reader, self.writer = await coroutine

    @connection
    async def readline(self) -> bytes:
        """
        Read and return one line, where "line" is a sequence
        of bytes ending with \\n.
        """

        return await self.reader.readline()

    @connection
    async def readuntil(self, separator: bytes = b"\n") -> bytes:
        """
        Read and return data from the stream until separator
        is found.
        """

        return await self.reader.readuntil(separator)

    @connection
    async def write(self, data: bytes) -> None:
        try:
            self.writer.write(data)
            await self.writer.drain()
        except ConnectionError:
            await self.close()
            raise

    @connection
    async def write_readline(self, data: bytes) -> bytes:
        """
        Write a message string and immediately return line
        from stream.
        """

        try:
            self.writer.write(data)
            await self.writer.drain()
        except ConnectionError:
            await self.close()
            raise
        return await self.reader.readline()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from asyncio import iscoroutine
from asyncio import StreamReaderProtocol
from typing import Callable


class CallbackProtocol(StreamReaderProtocol):
    """Callback protocol.

    A wrapper for StreamReaderProtocol that allows users to
    implement their own custom callback functions. Each user
    function will be called after to the
    StreamReaderProtocol helper method.
    """

    def _exec_callback(self, name: str, *args, **kwargs) -> None:
        callback = getattr(self, name, None)
        if callback is None:
            return
        result = callback(*args, **kwargs)
        if iscoroutine(result):
            self._loop.create_task(result)

    def connection_made(self, transport):
        result = super().connection_made(transport)
        self._exec_callback("connection_made_cb", transport)
        return result

    def connection_lost(self, exc):
        result = super().connection_lost(exc)
        self._exec_callback("connection_lost_cb", exc)
        return result

    def data_received(self, data):
        result = super().data_received(data)
        self._exec_callback("data_received_cb", data)
        return result

    def eof_received(self):
        result = super().eof_received()
        self._exec_callback("eof_received_cb")
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from asyncio import TimeoutError
from unittest import IsolatedAsyncioTestCase
from unittest import main
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import patch
from time import time

from uscpi.client import open_connection
from uscpi.client import TCP


class TestTCP(IsolatedAsyncioTestCase):
    def setUp(self):
        self.tcp = TCP("127.0.0.1", 8080)
        self.mock_reader = AsyncMock()
        self.mock_writer = AsyncMock()
        self.mock_writer.drain = AsyncMock()
        self.mock_writer.write = Mock()

    async def test_open_connection(self):
        response = await open_connection(
            "127.0.0.1", 8080, 65536, None, None, None, None
        )
        self.assertIsInstance(response, tuple)
        _, writer = response
        writer.close()
        await writer.wait_closed()

    async def test_tcp_timeout(self):
        tcp = TCP("10.0.0.0", 8080, 0.1)
        with self.assertRaises(TimeoutError):
            start = time()
            await tcp.open()
        delta = time() - start
        self.assertAlmostEqual(delta, 0.1, places=1)

    @patch("uscpi.client.open_connection")
    async def test_tcp_open(self, mock_open_connection):
        mock_open_connection.return_value = self.mock_reader, self.mock_writer
        await self.tcp.open()
        mock_open_connection.assert_called_once()
        self.assertIsNotNone(self.tcp.reader)
        self.assertIsNotNone(self.tcp.writer)

    @patch("uscpi.client.open_connection")
    async def test_tcp_close(self, mock_open_connection):
        self.tcp.reader = self.mock_reader
        self.tcp.writer = self.mock_writer
        await self.tcp.close()
        self.assertIsNone(self.tcp.reader)
        self.assertIsNone(self.tcp.writer)

    @patch("uscpi.client.open_connection")
    async def test_tcp_readline(self, mock_open_connection):
        mock_open_connection.return_value = self.mock_reader, self.mock_writer
        self.mock_reader.readline.return_value = b"test\n"
        self.assertEqual(await self.tcp.readline(), b"test\n")

    @patch("uscpi.client.open_connection")
    async def test_tcp_readuntil(self, mock_open_connection):
        mock_open_connection.return_value = self.mock_reader, self.mock_writer
        self.mock_reader.readuntil.return_value = b"test\n"
        self.assertEqual(await self.tcp.readuntil(b"\n"), b"test\n")

    @patch("uscpi.client.open_connection")
    async def test_tcp_write(self, mock_open_connection):
        mock_open_connection.return_value = self.mock_reader, self.mock_writer
        await self.tcp.write(b"test\n")
        self.mock_writer.drain.assert_awaited()
        self.mock_writer.write.assert_called_with(b"test\n")

    @patch("uscpi.client.open_connection")
    async def test_tcp_write_readline(self, mock_open_connection):
        mock_open_connection.return_value = self.mock_reader, self.mock_writer
        self.mock_reader.readline.return_value = b"test\n"
        response = await self.tcp.write_readline(b"test\n")
        self.mock_writer.drain.assert_awaited()
        self.mock_writer.write.assert_called_with(b"test\n")
        self.assertEqual(response, b"test\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest import main
from unittest.mock import Mock

from uscpi.protocol import CallbackProtocol


class TestProtocol(TestCase):
    def setUp(self):
        self.srp = CallbackProtocol(Mock(), loop=Mock())

    def test_callback_protocol_connection_made_cb(self):
        self.srp.connection_made_cb = Mock()
        self.srp.connection_made(transport=Mock())
        self.srp.connection_made_cb.assert_called_once()

    def test_callback_protocol_connection_lost_cb(self):
        self.srp.connection_lost_cb = Mock()
        self.srp.connection_lost(exc=Mock())
        self.srp.connection_lost_cb.assert_called_once()

    def test_callback_protocol_data_received_cb(self):
        self.srp.data_received_cb = Mock()
        self.srp.data_received(data=Mock())
        self.srp.data_received_cb.assert_called_once()

    def test_callback_protocol_eof_received_cb(self):
        self.srp.eof_received_cb = Mock()
        self.srp.eof_received()
        self.srp.eof_received_cb.assert_called_once()


if __name__ == "__main__":
    main()
